"""A reduced version of the Open Babel chemical toolbox to work with
ManipulateAggregates
"""

# Copyright (C) 2020 by Torsten Sachse
# All rights reserved.
#
# This file is part of MaAg-bel.
# The contents of this file are covered by the terms of the GPL v2 license,
# see the file COPYING in the root if the repository for the license.
#
# MaAg-bel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
#
# MaAg-bel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MaAg-bel. If not, see <http://www.gnu.org/licenses/>.

import os

# Set some environment variables needed by the library if not yet set
__MAABEL_PATH = os.path.dirname(os.path.realpath(__file__))
# The path to some data files needed by the forcefields
__DATA_PATH = os.path.join(__MAABEL_PATH, "data")
if os.environ.get("BABEL_DATADIR", None) is None:
    os.environ["BABEL_DATADIR"] = __DATA_PATH
# Import the C++ functionality
from .cpp import *



"""
pybel - A Cinfony module for accessing Open Babel

Global variables:
  ob - the underlying SWIG bindings for Open Babel
  informats - a dictionary of supported input formats
  outformats - a dictionary of supported output formats
  forcefields - a list of supported forcefields

Modified to be a part of MaAg-bel
"""

# Copyright (c) 2008-2012, Noel O'Boyle; 2012, Adria Cereto-Massague
# Copyright (C) 2020 by Torsten Sachse
# All rights reserved.
#
# This file is part of MaAg-bel.
# The contents of this file are covered by the terms of the GPL v2 license,
# see the file COPYING in the root if the repository for the license.
#
# MaAg-bel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
#
# MaAg-bel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MaAg-bel. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import logging

logger = logging.getLogger(__name__)

import maagbel as ob

_obfuncs = _obconsts = ob


def _formatstodict(list):
    broken = [
        x.replace("[Read-only]", "").replace("[Write-only]", "").split(" -- ")
        for x in list
    ]
    broken = [(x, y.strip()) for x, y in broken]
    return dict(broken)


def _getplugins(findplugin, names):
    return dict([(x, findplugin(x)) for x in names if findplugin(x)])


def getpluginconfigs(plugin):
    plugins = [_x.lower() for _x in _getpluginnames("plugins")]
    if plugin == "plugins":
        result = plugins
    elif plugin in plugins:
        result = [_x.lower() for _x in _getpluginnames(plugin)]
    else:
        raise ValueError(
            "Unknown plugin type '{}', known types are: {}".format(
                plugin, ", ".join(plugins)
            )
        )
    return result


def _getpluginnames(ptype):
    plugins = ob.vectorString()
    ob.OBPlugin.ListAsVector(ptype, None, plugins)
    return [x.split()[0] for x in plugins]


_obconv = ob.OBConversion()
_builder = ob.OBBuilder()

# A dictionary of supported input formats
informats = _formatstodict(_obconv.GetSupportedInputFormat())
# A dictionary of supported output formats
outformats = _formatstodict(_obconv.GetSupportedOutputFormat())

# A list of supported forcefields
forcefields = [_x.lower() for _x in _getpluginnames("forcefields")]
_forcefields = _getplugins(ob.OBForceField.FindType, forcefields)


def readfile(format, filename, opt=None):
    """Iterate over the molecules in a file.

    Args:
        format: see the informats variable for a list of available input formats
        filename: tha name of the file from which to read the molecular data
        opt: a dictionary of format-specific options For format options with no
            parameters, specify the value as None.

    You can access the first molecule in a file using the next() function:
    >>> mol = next(readfile("smi", "myfile.smi"))

    You can make a list of the molecules in a file using:
    >>> mols = list(readfile("smi", "myfile.smi"))

    You can iterate over the molecules in a file as shown in the following code snippet:
    >>> atomtotal = 0
    >>> for mol in readfile("sdf", "head.sdf"):
    >>>     atomtotal += len(mol.atoms)
    """
    if opt is None:
        opt = {}
    obconversion = ob.OBConversion()
    formatok = obconversion.SetInFormat(format)
    for k, v in opt.items():
        if v is None:
            obconversion.AddOption(k, obconversion.INOPTIONS)
        else:
            obconversion.AddOption(k, obconversion.INOPTIONS, str(v))
    if not formatok:
        raise ValueError("%s is not a recognised Open Babel format" % format)
    if not os.path.isfile(filename):
        raise IOError("No such file: '%s'" % filename)

    def filereader():
        obmol = ob.OBMol()
        notatend = obconversion.ReadFile(obmol, filename)
        while notatend:
            yield Molecule(obmol)
            obmol = ob.OBMol()
            notatend = obconversion.Read(obmol)

    return filereader()


class Outputfile(object):
    """Represent a file to which *output* is to be sent.

    Although it's possible to write a single molecule to a file by calling the write()
    method of a molecule, if multiple molecules are to be written to the same file you
    should use the Outputfile class.

    Args:
        format: see the outformats variable for a list of available output formats
        filename
        overwrite: if the output file already exists, should it be overwritten? (default
            is False)
        opt: a dictionary of format-specific options For format options with no
            parameters, specify the value as None.

    Methods:
       write(molecule)
       close()
    """

    def __init__(self, format, filename, overwrite=False, opt=None):
        if opt is None:
            opt = {}
        self.format = format
        self.filename = filename
        if not overwrite and os.path.isfile(self.filename):
            raise IOError(
                "%s already exists. Use 'overwrite=True' to overwrite it."
                % self.filename
            )

        self.obConversion = ob.OBConversion()
        formatok = self.obConversion.SetOutFormat(self.format)
        if not formatok:
            raise ValueError("%s is not a recognised Open Babel format" % format)
        if filename and filename.split(".")[-1] == "gz":
            self.obConversion.AddOption("z", self.obConversion.GENOPTIONS)
        for k, v in opt.items():
            if v is None:
                self.obConversion.AddOption(k, self.obConversion.OUTOPTIONS)
            else:
                self.obConversion.AddOption(k, self.obConversion.OUTOPTIONS, str(v))
        self.total = 0  # The total number of molecules written to the file

    def write(self, molecule):
        """Write a molecule to the output file.

        Args:
            molecule
        """
        if not self.filename:
            raise IOError("Outputfile instance is closed.")

        if self.total == 0:
            self.obConversion.WriteFile(molecule.OBMol, self.filename)
        else:
            self.obConversion.Write(molecule.OBMol)
        self.total += 1

    def close(self):
        """Close the Outputfile to further writing."""
        self.obConversion.CloseOutFile()
        self.filename = None


class Molecule(object):
    """Represent a Pybel Molecule.

    Args:
        OBMol: an Open Babel OBMol

    Methods:
       addh(), localopt(), make3D(), removeh(), write()

    The underlying Open Babel molecule can be accessed using the attribute:
       OBMol
    """

    def __init__(self, OBMol):
        self.OBMol = OBMol

    @property
    def charge(self):
        return self.OBMol.GetTotalCharge()

    @property
    def conformers(self):
        return self.OBMol.GetConformers()

    @property
    def dim(self):
        return self.OBMol.GetDimension()

    @property
    def energy(self):
        return self.OBMol.GetEnergy()

    @property
    def exactmass(self):
        return self.OBMol.GetExactMass()

    @property
    def formula(self):
        return self.OBMol.GetFormula()

    @property
    def molwt(self):
        return self.OBMol.GetMolWt()

    @property
    def spin(self):
        return self.OBMol.GetTotalSpinMultiplicity()

    def _gettitle(self):
        return self.OBMol.GetTitle()

    def _settitle(self, val):
        self.OBMol.SetTitle(val)

    title = property(_gettitle, _settitle)

    def write(self, format="smi", filename=None, overwrite=False, opt=None):
        """Write the molecule to a file or return a string.

        If a filename is specified, the result is written to a file. Otherwise, a string
        is returned containing the result.

        To write multiple molecules to the same file you should use the Outputfile
        class.

        Args:
            format: see the informats variable for a list of available output formats
                (default is "smi")
            filename: default is None
            overwite: if the output file already exists, should it be overwritten?
                (default is False)
            opt: a dictionary of format specific options For format options with no
                parameters, specify the value as None.

        """
        if opt is None:
            opt = {}
        obconversion = ob.OBConversion()
        formatok = obconversion.SetOutFormat(format)
        if not formatok:
            raise ValueError("%s is not a recognised Open Babel format" % format)
        if filename and filename.split(".")[-1] == "gz":
            obconversion.AddOption("z", self.obConversion.GENOPTIONS)
        for k, v in opt.items():
            if v is None:
                obconversion.AddOption(k, obconversion.OUTOPTIONS)
            else:
                obconversion.AddOption(k, obconversion.OUTOPTIONS, str(v))

        if filename:
            if not overwrite and os.path.isfile(filename):
                raise IOError(
                    ("%s already exists. Use 'overwrite=True' to " "overwrite it.")
                    % filename
                )
            obconversion.WriteFile(self.OBMol, filename)
            obconversion.CloseOutFile()
        else:
            return obconversion.WriteString(self.OBMol)

    def localopt(self, forcefield="mmff94", steps=500):
        """Locally optimize the coordinates.

        If the molecule does not have any coordinates, make3D() is called before the
        optimization. Note that the molecule needs to have explicit hydrogens. If not,
        call addh().

        Args:
            forcefield: default is "mmff94". See the forcefields variable for a list of
                available forcefields.
            steps: default is 500
        """
        forcefield = forcefield.lower()
        if self.dim != 3:
            self.make3D(forcefield)
        ff = _forcefields[forcefield]
        success = ff.Setup(self.OBMol)
        if not success:
            raise RuntimeError("Local optimization with forcefield unsuccessful")
        ff.SteepestDescent(steps)
        ff.GetCoordinates(self.OBMol)

    def make3D(self, forcefield="mmff94", steps=50):
        """Generate 3D coordinates.

        Once coordinates are generated, hydrogens are added and a quick local
        optimization is carried out with 50 steps and the MMFF94 forcefield. Call
        localopt() if you want to improve the coordinates further.

        Args:
            forcefield: default is "mmff94". See the forcefields variable for a list of
                available forcefields.
            steps: default is 50
        """
        forcefield = forcefield.lower()
        _builder.Build(self.OBMol)
        self.addh()
        self.localopt(forcefield, steps)

    def addh(self):
        """Add hydrogens."""
        self.OBMol.AddHydrogens()

    def removeh(self):
        """Remove hydrogens."""
        self.OBMol.DeleteHydrogens()

    def __str__(self):
        return self.write()



import os
import itertools
from subprocess import Popen, PIPE
from setuptools import setup, Extension, find_packages
from distutils.command.build import build as build_orig

# Misc executables
UNAME = "uname"


def get_long_description():
    """Load the contents of README.md to use as a long description
    
    Returns:
        The content of README.md as a string.
    """
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
    return long_description


# Error classes
class SetupError(RuntimeError):
    """Base error class for this setup file"""

    pass


class CommandNotFoundError(SetupError):
    """Raised when running a command cannot be found"""

    pass


class CannotRunCommandError(SetupError):
    """Raised when running a command returns with an error"""

    pass


class DependencyNotFoundError(SetupError):
    """Raised when a dependency cannot be found"""

    pass


# Helper functions
def call_prog(prog, args, encoding=None, strip=False):
    """Call an executable and return its standard output
    Args:
        prog: (str) - the executable's name
        args: (list of str) - arguments for the executable
        encoding: (str, optional, default: None) - chose an encoding to convert the
            output of called programs to Python strings, uses the system-default
            encoding by default
        strip: (bool, optional, default: False) - whether to strip newlines and spaces
            from the end of the returned string

    Returns:
        The standard output of the executable called with the given arguments.

    Raises:
        CommandNotFoundError if the executable cannot be found.
        CannotRunCommandError if the executable did not yield an exit code of zero.
    """
    try:
        process = Popen([prog] + args, stdout=PIPE)
    except FileNotFoundError:
        raise CommandNotFoundError("Command '{}' not found".format(prog))
    (output, err) = process.communicate()
    exit_code = process.wait()
    if exit_code != 0:
        raise CannotRunCommandError(
            "Error running '{} {}', error: {}".format(prog, " ".join(args), err)
        )
    else:
        if encoding is not None:
            output = output.decode(encoding=encoding)
        else:
            output = output.decode()
        if strip:
            output = output.rstrip()
        return output


def get_lib_dirs(kind):
    """Obtain library dirs depending on the architecture.

    Supports determining system library dirs (kind="system") and conda library dirs
    (kind="conda").

    This may not be necessary depending on the compiler used. If the program "uname" is
    not installed, the system's architecture cannot be determined and a standard set of
    directories is returned.

    Args:
        kind: (string) - the kind of library paths you want

    Returns:
        A list of strings containing the library paths of the specified type.
    """
    if kind == "system":
        dirs = ["/usr/lib/"]
        try:
            arch = call_prog(UNAME, ["-m"], strip=True)
        except CommandNotFoundError:
            arch = None
        if arch is not None:
            dirs.append("/usr/lib/{}-linux-gnu/".format(arch))
    elif kind == "conda":
        dirs = []
        prefix = get_conda_prefix()
        if prefix is not None:
            dirs.append(os.path.join(prefix, "lib"))
    else:
        raise ValueError("Unknown kind '{}'".format(kind))
    return dirs


def get_include_dirs(kind):
    """Obtain include dirs depending on the architecture.

    Supports determining system include dirs (kind="system") and conda include dirs
    (kind="conda").

    This may not be necessary depending on the compiler used. If the program "uname" is
    not installed, the system's architecture cannot be determined and a standard set of
    directories is returned.

    Args:
        kind: (string) - the kind of include paths you want

    Returns:
        A list of strings containing the include paths of the specified type.
    """
    if kind == "system":
        dirs = ["/usr/include/"]
        try:
            arch = call_prog(UNAME, ["-m"], strip=True)
        except CommandNotFoundError:
            arch = None
        if arch is not None:
            dirs.append("/usr/include/{}-linux-gnu/".format(arch))
    elif kind == "conda":
        dirs = []
        prefix = get_conda_prefix()
        if prefix is not None:
            dirs.append(os.path.join(prefix, "include"))
    else:
        raise ValueError("Unknown kind '{}'".format(kind))
    return dirs


def get_eigen_include_dir():
    """Try to find the eigen include directory.

    Search all directories in the system and conda include paths for a directory called
    "eigen3". Stop at the first one found and return its path.

    Returns:
        A path to the eigen3 include directory if found, None otherwise
    """
    # Allow explicitly setting the eigen include directory
    env_vars = [
        "MAINST_EIGEN3_DIR",
    ]
    for var in env_vars:
        eigen_dir = os.environ.get(var, None)
        if eigen_dir is not None:
            return eigen_dir

    # Unless otherwise specified, prefer the conda installation of eigen3. If not found,
    # search the system paths.
    kinds = ("conda", "system")
    env_vars = [
        "MAINST_EIGEN3_PREFER_SYSTEM",
    ]
    for var in env_vars:
        if os.environ.get(var, "0") == "1":
            kinds = ("system", "conda")
    for kind in kinds:
        for d in get_include_dirs(kind):
            content = os.listdir(d)
            for c in content:
                if c == "eigen3" and os.path.isdir(d):
                    return os.path.join(d, c)

    return None


def get_conda_prefix():
    """Determine whether a conda environment is active and return its path
    
    Returns:
        The path if a conda env is active and None otherwise.
    """
    path = os.environ.get("CONDA_PREFIX", None)
    return path


class build(build_orig):
    """A class that allows re-ordering the build process.

    Re-order the commands build_py and build_ext so that build_ext is run first. That
    way, swig-generated Python files will be available for the build_py step.

    Taken from https://stackoverflow.com/questions/50239473/building-a-module-with-setuptools-and-swig
    """

    def finalize_options(self):
        super().finalize_options()
        condition = lambda l: l[0] == "build_ext"
        t1, t2 = itertools.tee(self.sub_commands)
        rest, sub_build_ext = (
            itertools.filterfalse(condition, t1),
            filter(condition, t2),
        )
        self.sub_commands[:] = list(sub_build_ext) + list(rest)


def is_file_type(file, type):
    """Determine whether a file is of a certain type.

    Supported types are "source" and "header". Files ending in ".cpp" or ".c" are
    considered "source" files, whereas files ending in ".hpp" or ".h" are considered
    "header" files.

    Args:
        file: (string) - the name of a file to be checked
        type: (string) - one of the supported file types

    Raises:
        ValueError if the type of file is not known
    """
    if type == "source":
        matches = file.endswith(".cpp") or file.endswith(".c")
    elif type == "header":
        matches = file.endswith(".hpp") or file.endswith(".h")
    else:
        raise ValueError("Unsupported file type '{}'".format(type))
    return matches


def get_files_in_dir(dir):
    """Return a list of relative paths to all files in a directory recursively.

    Args:
        dir: (string) - path to a directory
    """
    result = []
    for root, __, files in os.walk(os.path.join(dir)):
        result += [os.path.join(root, f) for f in files]
    return result


def get_ext_modules(basename):
    """Get a list of external modules.

    This will install the OpenBabel C++ module with reduced functionality.

    Args:
        basename: (str) - the name of the main module, used to ensure that any
            extensions are sub-modules

    Returns:
        A list of extensions to be passed to the ext_modules argument of the setup
        function.
    """

    # Determine source files
    files = get_files_in_dir(os.path.join(os.path.curdir, "src"))
    sources = [f for f in files if is_file_type(f, "source")]
    sources.append("maagbel/cpp.i")

    # # Determine header files
    # files += get_files_in_dir(os.path.join(os.path.curdir, "include"))
    # headers = [f for f in files if is_file_type(f, "header")]

    libraries = ["pthread", "c", "m"]
    extra_compile_args = [
        "-std=c++14",
        "-v",
        "-pedantic",
        "-Wall",
        "-Wextra",
        "-fPIC",
        "-O3",
    ]
    swig_opts = ["-c++", "-Iinclude"]

    include_dirs = get_include_dirs("system") + get_include_dirs("conda")
    eigen_include = get_eigen_include_dir()
    if eigen_include is not None:
        include_dirs.append(eigen_include)
    include_dirs.append("include")

    library_dirs = get_lib_dirs("system") + get_lib_dirs("conda")

    # Create the main C++ extension with either min or max funtionality
    maagbel_cpp_ext = Extension(
        basename + "._cpp",
        sources=sources,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        language="c++",
        swig_opts=swig_opts,
        extra_compile_args=extra_compile_args,
    )
    ext_modules = [maagbel_cpp_ext]

    return ext_modules


def check_dependencies():
    """Check whether all required dependencies can be found. Raise an exception if one
    is missing.
    """

    # Check whether SWIG is installed and can be found
    try:
        call_prog("swig", ["-version"])
    except CommandNotFoundError:
        raise CommandNotFoundError(
            "Cannot find SWIG and execute, please install it on your system, "
            "run 'swig --version' to check whether swig works"
        )
    # Check whether Eigen3 is installed and can be found
    if get_eigen_include_dir() is None:
        raise DependencyNotFoundError(
            "Cannot find installation of Eigen3, please install it on your system"
        )


def main():
    # Package data
    PACKAGE_DATA = {
        "name": "maagbel",
        "version": "0.1.2",
        "description": "Reduced fork of OpenBabel for the ManipulateAggregates tools",
        "author": "Torsten Sachse",
        "mail": "torsten.sachse@gmail.com",
        "url": "https://github.com/razziel89/MaAg-bel",
        "long_description": get_long_description(),
        "long_description_content_type": "text/markdown",
        "classifiers": [
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
            "Natural Language :: English",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: C++",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3 :: Only",
        ],
    }

    check_dependencies()

    setup(
        ext_modules=get_ext_modules(PACKAGE_DATA["name"]),
        packages=find_packages(),
        cmdclass={"build": build},
        package_data={PACKAGE_DATA["name"]: ["data/*"]},
        **PACKAGE_DATA,
    )


if __name__ == "__main__":
    main()
